@@ -7,6 +7,7 @@ use std::str::FromStr;
 use crate::errors::StdError;
 
 use super::Fraction;
+use super::Isqrt;
 use super::Uint128;
 
 /// A fixed-point decimal value with 18 fractional digits, i.e. Decimal(1_000_000_000_000_000_000) == 1.0
@@ -55,6 +56,32 @@ impl Decimal {
     pub fn is_zero(&self) -> bool {
         self.0 == 0
     }
+
+    /// Returns the approximate square root as a Decimal.
+    ///
+    /// This should not overflow or panic.
+    pub fn sqrt(&self) -> Self {
+        // Algorithm described in https://hackmd.io/@webmaster128/SJThlukj_
+        // We start with the highest precision possible and lower it until
+        // there's no overflow.
+        (0..=9)
+            .rev()
+            .find_map(|i| self.sqrt_with_precision(i))
+            // The last step (i = 0) is guaranteed to succeed because `isqrt(u128::MAX) * 10^9` does not overflow
+            .unwrap()
+    }
+
+    /// Lower precision means more aggressive rounding, but less risk of overflow.
+    /// Precision *must* be a number between 0 and 9 (inclusive).
+    ///
+    /// Returns `None` if the internal multiplication overflows.
+    fn sqrt_with_precision(&self, precision: u32) -> Option<Self> {
+        let inner_mul = 100u128.pow(precision);
+        self.0.checked_mul(inner_mul).map(|inner| {
+            let outer_mul = 10u128.pow(9 - precision);
+            Decimal(inner.isqrt() * outer_mul)
+        })
+    }
 }
 
 impl Fraction<u128> for Decimal {
@@ -150,6 +177,14 @@ impl ops::Add for Decimal {
     }
 }
 
+impl ops::Sub for Decimal {
+    type Output = Self;
+
+    fn sub(self, other: Self) -> Self {
+        Decimal(self.0 - other.0)
+    }
+}
+
 /// Both d*u and u*d with d: Decimal and u: Uint128 returns an Uint128. There is no
 /// specific reason for this decision other than the initial use cases we have. If you
 /// need a Decimal result for the same calculation, use Decimal(d*u) or Decimal(u*d).
@@ -174,6 +209,20 @@ impl ops::Mul<Uint128> for Decimal {
     }
 }
 
+impl ops::Div<Uint128> for Decimal {
+    type Output = Self;
+
+    fn div(self, rhs: Uint128) -> Self::Output {
+        Decimal(self.0 / rhs.u128())
+    }
+}
+
+impl ops::DivAssign<Uint128> for Decimal {
+    fn div_assign(&mut self, rhs: Uint128) {
+        self.0 /= rhs.u128();
+    }
+}
+
 /// Serializes as a decimal string
 impl Serialize for Decimal {
     fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
@@ -489,6 +538,24 @@ mod tests {
         assert_eq!(value.0, DECIMAL_FRACTIONAL * 3 / 2);
     }
 
+    #[test]
+    #[should_panic(expected = "attempt to add with overflow")]
+    fn decimal_add_overflow_panics() {
+        let _value = Decimal::MAX + Decimal::percent(50);
+    }
+
+    #[test]
+    fn decimal_sub() {
+        let value = Decimal::one() - Decimal::percent(50); // 0.5
+        assert_eq!(value.0, DECIMAL_FRACTIONAL / 2);
+    }
+
+    #[test]
+    #[should_panic(expected = "attempt to subtract with overflow")]
+    fn decimal_sub_overflow_panics() {
+        let _value = Decimal::zero() - Decimal::percent(50);
+    }
+
     #[test]
     // in this test the Decimal is on the right
     fn uint128_decimal_multiply() {
@@ -527,6 +594,84 @@ mod tests {
         assert_eq!(left * right, Uint128(0));
     }
 
+    #[test]
+    fn decimal_uint128_division() {
+        // a/b
+        let left = Decimal::percent(150); // 1.5
+        let right = Uint128(3);
+        assert_eq!(left / right, Decimal::percent(50));
+
+        // 0/a
+        let left = Decimal::zero();
+        let right = Uint128(300);
+        assert_eq!(left / right, Decimal::zero());
+    }
+
+    #[test]
+    #[should_panic(expected = "attempt to divide by zero")]
+    fn decimal_uint128_divide_by_zero() {
+        let left = Decimal::percent(150); // 1.5
+        let right = Uint128(0);
+        let _result = left / right;
+    }
+
+    #[test]
+    fn decimal_uint128_div_assign() {
+        // a/b
+        let mut dec = Decimal::percent(150); // 1.5
+        dec /= Uint128(3);
+        assert_eq!(dec, Decimal::percent(50));
+
+        // 0/a
+        let mut dec = Decimal::zero();
+        dec /= Uint128(300);
+        assert_eq!(dec, Decimal::zero());
+    }
+
+    #[test]
+    #[should_panic(expected = "attempt to divide by zero")]
+    fn decimal_uint128_div_assign_by_zero() {
+        // a/0
+        let mut dec = Decimal::percent(50);
+        dec /= Uint128(0);
+    }
+
+    #[test]
+    fn decimal_uint128_sqrt() {
+        assert_eq!(Decimal::percent(900).sqrt(), Decimal::percent(300));
+
+        assert!(Decimal::percent(316) < Decimal::percent(1000).sqrt());
+        assert!(Decimal::percent(1000).sqrt() < Decimal::percent(317));
+    }
+
+    /// sqrt(2) is an irrational number, i.e. all 18 decimal places should be used.
+    #[test]
+    fn decimal_uint128_sqrt_is_precise() {
+        assert_eq!(
+            Decimal::from_str("2").unwrap().sqrt(),
+            Decimal::from_str("1.414213562373095048").unwrap() // https://www.wolframalpha.com/input/?i=sqrt%282%29
+        );
+    }
+
+    #[test]
+    fn decimal_uint128_sqrt_does_not_overflow() {
+        assert_eq!(
+            Decimal::from_str("400").unwrap().sqrt(),
+            Decimal::from_str("20").unwrap()
+        );
+    }
+
+    #[test]
+    fn decimal_uint128_sqrt_intermediate_precision_used() {
+        assert_eq!(
+            Decimal::from_str("400001").unwrap().sqrt(),
+            // The last two digits (27) are truncated below due to the algorithm
+            // we use. Larger numbers will cause less precision.
+            // https://www.wolframalpha.com/input/?i=sqrt%28400001%29
+            Decimal::from_str("632.456322602596803200").unwrap()
+        );
+    }
+
     #[test]
     fn decimal_to_string() {
         // Integers
